{-

SPS/APS - Security Protocol Specification langauge
Copyright Omar Saad Almousa, Technical Univerity of Denmark (DTU) 2013-2015

All Rights Reserved.

-}

module GeneratePlainStrands where -- 
import StrandAST
import Msg
import Ast
import System.Environment
import Parser
import Lexer
----------------------------------------------------------
--generateStrand:: [peer]-> Protocol -> [Strand]
-- It takes a list of peers (Agents) and returns A list of strands
------------------------------------------------------------
-- rename getChannel to getDetails
-- This function checks whether the actions of a protocol follow token-passing 
chkTokenPassing (a:rest) = chkTokenPassing0 2 (getRecver a) rest
chkTokenPassing0 n p [] = True
chkTokenPassing0 n p (a:rest) = if(p==(getSender a)) then chkTokenPassing0  (n+1) (getRecver a) rest
                                else error("The protocol must follow token-passing form, i.e., the receiver of a message must be the sender of the next one. Check Actions section, message sending # " ++ show(n))
------------------------------------------------------------
excOrder all@(a:rest) =  if(isExchange a) then (getSender a): excOrder0 all else excOrder rest
excOrder [] = []
excOrder0 (a:rest) =  if(isExchange a) then (getRecver a): excOrder0 rest else excOrder0 rest
excOrder0 [] = []
------------------------------------------------------------
excOrdProt =  excOrder . getActions
------------------------------------------------------------------------
generateStrands :: Protocol -> [Strand]
generateStrands protocol = if chkTokenPassing (getActions protocol) then
                                  generateStrand (onlySenders (getActions protocol)) protocol else
                                  error "The protocol is not token-passing"
-------------------------------------------------------------
-- isVar && not elem lets
getEmbdFresh:: Protocol-> Protocol
getEmbdFresh p@(nm,tp,mp,fr,(knows, cnds),mc,ac,gl,pr, rs, ab) = p
-- getEmbdFresh p@(nm,tp,mp,fr,(knows, cnds),mc,ac,gl,ab) = (nm,tp,mp,fr,((embdTo knows (getFreshActs [] ac (getRoles tp))), cnds),mc,ac,gl,ab)
-------------------------------------------------------------
--embdTo: takes knowledges, and a list of pairs in the form (agent, freshValue) 
-- then it embeds to each knowledge the corresponding fresh values
embdTo:: [Aknowledge] -> [(Ident, [Msg])] -> [Aknowledge]
embdTo ((id, know):rest) pairs= (id, know ++ concat (freshOfAgent id pairs)) : (embdTo rest pairs)
embdTo [] _=[]
-------------------------------------------------------------filter (not (elem taken)) 
getFreshActs taken (act@(Exchange a@(ch,m1,m2,m3)):rest) roles= (getSender act, getFreshMsg  roles taken m1): 
                                                    (getFreshActs ((getFreshMsg  roles taken m1)++taken) rest  roles)
getFreshActs taken (a:rest) roles=     getFreshActs taken rest roles
getFreshActs _ [] _ = []
-------------------------------------------------------------
-- ==========TAKEN list, to denote that a afresh value is already created by some other agent!
getFreshMsg  roles taken m@(Atom x)  = if((isFresh x roles) && (not (elem m taken))) then [m] else []
getFreshMsg  roles taken m@(Comp Apply opt lst) = concat (map (getFreshMsg  roles taken) (tail lst))
getFreshMsg  roles taken m@(Comp op opt lst) = concat (map (getFreshMsg  roles taken) (lst))
-------------------------------------------------------------
freshOfAgent agent fresh = [y| (x,y)<- fresh, x==agent]
-------------------------------------------------------------
chkNoFrshAll:: Protocol->Protocol
chkNoFrshAll p@(nm,tp,mp,fr,(idknows,cnds),mc,ac,gl,pr,rs,ab)= if (idknows==[]) then error("Do not we have any knowledges?")
                                                        else(nm,tp,mp,fr,(chkNoFrshAll0 idknows (getRoles tp),cnds) ,mc,ac,gl,pr,rs,ab) 
-------------------------------------------------------------
chkNoFrshAll0  ((id,know):rest) roles      =  (id, (chkNoFrshOne id know roles)) : (chkNoFrshAll0  rest roles)
chkNoFrshAll0  [] roles                 = []
-------------------------------------------------------
chkNoFrshOne id (m@(Atom x):rest) roles = if(isFresh x roles) 
                                                then error("Knowledge section cannot contain fresh variables. Check the message: " ++ show(x) ++ " in agent " ++ show(id)++ "'s knowledge")
                                            else m: (chkNoFrshOne id rest roles)
    
chkNoFrshOne id (m@(Comp op opt lst):rest) roles = m: (chkNoFrshOne id rest roles)
chkNoFrshOne id [] roles                 = []
-------------------------------------------------------
-------------------------------------------------------
isFresh x roles= (isVar x) && (not (elem x roles))
isVar x = elem (head x) ['A'..'Z']
isNotVar x = not (isVar x)
getRoles ((Agent _ _ _,agents):rest) = filter isVar agents
getRoles (t:rest)= getRoles rest
getRoles [] = error("Do not we have any agents in this protocol. Check Types section?")
-------------------------------------------------------
-------------------------------------------------------------
generateStrand [] _= []
generateStrand (x:xs) protocol@(pname, _, _, _, pknowl,_, pactions, pgoals,pr,rs,_) = 
        (Strand {owner = x, name = pname, knowledge = labelKnowledge (getAgentKnowledge x pknowl), 
         rest = (generateRest x pgoals pactions pgoals)}) 
                    : (generateStrand xs protocol)
------------------------------------------------------------------------------
getActions protocol@(_, _, _, _, _,_, pactions, _,pr,rs,_)  = pactions
-------------------------------------------------
-- only senders
onlySenders actions = rmvdups ([getSender act| act<-actions, isExchange act] ++ [getRecver act| act<-actions, isExchange act]) --
isExchange (Exchange _)= True
isExchange  _= False
--------------------------
--------------------------
getAgentKnowledge a k= head (map snd (filter ((==a).fst) (getKnowledgeOnly(k))))
getKnowledgeOnly x = fst(x)
getConditionsOnly x = snd(x)
---------------------------------------------------------------------------
--Labelling messages stupidly!
createLabel (a,b)= a ++ show(b)
labelKnowledge [] = []
labelKnowledge x = zip (map Atom (map createLabel (zip (take (length x) (repeat "X")) (take (length x) [1..])))) x                
--------------------------------------------------------
-- CreateFresh (Channel, (Type, [Ident]))
--(Exchange a@(ch,m1,m2,m3))
getSender (Exchange (((a,_,_,_),_,_), _, _,_))= a
getSender (CreateFresh(((a,_,_,_),_,_),(_,_))) = a
getRecver (Exchange(( _,_,(b,_,_,_)), _, _,_)) = b
getRecver (CreateFresh(( _,_,(b,_,_,_)), (_,_))) = b
getMsg (Exchange(_, m, _,_)) = m
getMsg (CreateFresh _ )= error("CreateFresh has no Message")
-----------------------------------------------------------
getChSender (a,_,_) = a
getChRecver ( _,_,b)= b
-- getMsg (_, m, _,_) = m
----------------------------------------------------------
topGoals p f@(Secret msg peers)            = Connects
topGoals p f@(ChGoal ch msg)                = if(p == p1) then (Add Witness (ms:"witness":prsStr) msg prsMsg)
                                            else Connects
                                            where p1= getChSender ch;
                                                  p2 = getChRecver ch;
                                                  prsMsg=map peerToMsg [p1,p2];
                                                  prsStr=map peerToStr [p1,p2];
                                                  ms= getIdent msg
topGoals p f@(WAuthentication p1 p2 msg)    = if(p == p1) then (Add Witness (ms:"witness":prsStr) msg prsMsg) 
                                            else Connects
                                            where prsMsg=map peerToMsg [p1,p2];
                                                  prsStr=map peerToStr [p1,p2];
                                                  ms= getIdent msg
topGoals p f@(Authentication p1 p2 msg)    = if(p == p1) then (Add Witness (ms:"witness":prsStr) msg prsMsg)
                                            else Connects
                                            where prsMsg=map peerToMsg [p1,p2];
                                                  prsStr=map peerToStr [p1,p2];
                                                  ms= getIdent msg
------------------------------------------------------------------------------------------------------------------------------------------
-- botGoals -- generates the goal events that must be appended at the end of the process macro
botGoals p f@(Secret msg peers)        = if(elem p peers) then (Add Secrt ([ms,"secret"]++prsStr) msg prsMsg)
                                                            else Connects
                                                    where prsMsg=map peerToMsg peers;
                                                          prsStr=map peerToStr peers;
                                                          ms= getIdent msg
botGoals p f@(ChGoal ch msg) = if(p == p2) then (Add Wrequest  (ms:"wrequest":prsStr) msg prsMsg)
                                else Connects
                                where p1= getChSender ch; 
                                      p2 = getChRecver ch;
                                      prsMsg=map peerToMsg [p1,p2];
                                      prsStr=map peerToStr [p1,p2];
                                      ms= getIdent msg
botGoals p f@(WAuthentication p1 p2 msg)= if(p == p2) then (Add Wrequest (ms:"wrequest":prsStr) msg prsMsg)
                                            else Connects
                                            where prsMsg=map peerToMsg [p1,p2];
                                                  prsStr=map peerToStr [p1,p2];
                                                  ms= getIdent msg
botGoals p f@(Authentication p1 p2 msg)    = if(p == p2) then (Add Request (ms:"wrequest":prsStr) msg prsMsg)
                                            else Connects
                                            where prsMsg=map peerToMsg [p1,p2];
                                                  prsStr=map peerToStr [p1,p2];
                                                  ms= getIdent msg
----------------------------------------------------------------------------------------------------------------------
-- keyStr: generate a key string for the event (fact), the integer is to distinguish request(an wrequest) from witness in ChGoal, i.e., 0 for witness, 1 otherwise
keyStr n f@(Secret msg peers)        = (getIdent msg)++ concat( map peerToStr peers)
keyStr n f@(ChGoal ch msg)            = if(n==1) then ((getIdent msg) ++ peerToStr p1 ++ peerToStr p2)
                                        else ((getIdent msg) ++ peerToStr p1 ++ peerToStr p2)
                                        where p1= getChSender ch; 
                                              p2 = getChRecver ch
keyStr n f@(WAuthentication p1 p2 msg)    = if(n==1) then ((getIdent msg) ++ peerToStr p1 ++ peerToStr p2)    else ((getIdent msg) ++  peerToStr p1 ++ peerToStr p1)
keyStr n f@(Authentication p1 p2 msg)    = if(n==1) then ((getIdent msg) ++  peerToStr p1 ++ peerToStr p2)    else ((getIdent msg) ++ peerToStr p1 ++ peerToStr p1)
-----------------------------------------------------------------------------------------------
generateRest p (g:oals) acts goals  = (topGoals (toPeer p) g)     (generateRest p oals acts goals)
generateRest p [] (a:cts) goals     = (tr_s p a)                  (generateRest p [] cts goals)
generateRest p [] [] (g:oals)       = (botGoals (toPeer p) g)     (generateRest p [] [] oals)
generateRest p [] [] []             = EOS "SessionID"
-----------------------------------------------------------------------------------------------
-- converts an ident (string) to a peer structure!
-- toPeer::Ident->Peer
toPeer p = (p,Atom p, False, Nothing)
-- WORKING Version
-- generateRest:: String->Actions ->Strand0
-- generateRest peer [] goals     = generateGoals (peer, False, Nothing) goals
-- generateRest peer (l:ls) goals = (tr_s peer l) (generateRest peer ls goals)
--------------------------------------------
{-
-- generateGoals::this also generates a strand of all goals, obselete, since tit was replaced by botGoals and topGoals
generateGoals peer ((Secret msg peers):rest)            = if(elem peer peers) then (Add Secrt msg (map peerToMsg peers))(generateGoals peer rest) 
                                                            else (generateGoals peer rest) 
generateGoals peer ((ChGoal ch msg):rest)                = if(peer == p2) then (Add Wrequest msg (map peerToMsg [p1,p2]))(generateGoals peer rest)
                                                            else if(peer == p1) then (Add Witness msg (map peerToMsg [p1,p2]))(generateGoals peer rest)
                                                            else (generateGoals peer rest) 
                                                                where p1= getChSender ch; 
                                                                    p2 = getChRecver ch
generateGoals peer ((WAuthentication p1 p2 msg):rest)    = if(peer == p2) then (Add Wrequest msg (map peerToMsg [p1,p2]))(generateGoals peer rest)
                                                            else if(peer == p1) then (Add Witness msg (map peerToMsg [p1,p2]))(generateGoals peer rest)
                                                            else (generateGoals peer rest)
generateGoals peer ((Authentication p1 p2 msg):rest)    = if(peer == p2) then (Add Request msg (map peerToMsg [p1,p2]))(generateGoals peer rest)
                                                            else if(peer == p1) then (Add Witness msg (map peerToMsg [p1,p2]))(generateGoals peer rest)
                                                            else (generateGoals peer rest)
generateGoals peer [] = EOS "SessionID"
--     Witness     |     Request        |     Wrequest    |     Secret     |     Contains 
-}
-----------------------------------------
-- This converts peers to agent names as meesages i.e., the peer( "A", False, Nothing) is converted to Atom "A"


peerToMsg:: Peer->Msg
peerToMsg (a, b, c, d )= if(b==Atom a) then b else error("ERROR: Channel inconsistent!\n\n")
peerToStr:: Peer->String
peerToStr (a,b,c,d)= a
--tr_s::Action->a
tr_s p a@(Exchange x) =     if p == (getSender a) then (Snd x (getMsg a))
                        else if p == (getRecver a) then (Rcv x (getMsg a))
                        else Connects
tr_s p (CreateFresh (((a,_,_,_),_,_),(t,msgs)))= if p == a then (Fresh t (map Atom msgs) msgs) else Connects
-- tr_s p a@(CreateFresh x)= error("Hello Freshy!") -- Fresh t (map Atom msgs)
                        -- else error("Exchange")
                        -- else Omar "Nothing"
--tr_s p (Let x)= StrandAST.Nothing
-- tr_s _ _= Omar "How this appeared!"
tr_s _ _= error("Not Exchange")
---------------------------------------------------------------
rmvdups x= removeDups x []
removeDups [] sofar = []
removeDups (x:rest) sofar
     | elem x sofar = (removeDups rest sofar)
     | otherwise      = x:(removeDups rest (x:sofar))
--------------------------------
combineAll [] = Connects -- \s -> s --EOS "Hello"
combineAll (x:xs)= (x) . (combineAll xs)
---------------------------------------------------
getIdent (Atom ('f':'r':'m':rst))= rst
getIdent (Atom ('m':'p':'p':rst))= rst
getIdent (Atom x)= x
getIdent (Comp Apply opt ((Atom ('f':'r':'m':rst)):xs))= rst++"(" ++(getIdents xs)++")"
getIdent (Comp Apply opt (x:xs))= (getIdent x) ++ "(" ++ (getIdents xs)++")"
getIdent (Comp Cat opt ls)= "(" ++ (getIdents ls)++")"
getIdent (Comp op opt ls)= lowerize (show(op)) ++ "(" ++ (getIdents ls)++")"
----------------------
getIdents []=" "
getIdents [x]= lowerize (getIdent x)
getIdents (x:y:z)= lowerize (getIdent x) ++ ", " ++ (getIdents(y:z))
--------------------------------------------------------
owners x=[owner(a)| a<-x]
knowledges x=[knowledge(a)| a<-x]
--------------------------------------------------------------------------